package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.config.ApplicationProperties;
import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.dataformat.yaml.YAMLGenerator;
import net.minidev.json.JSONObject;
import net.minidev.json.JSONStyle;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.lib.Repository;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import java.io.*;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

@Service
public class ExerciseService {

    private final Logger log = LoggerFactory.getLogger(ExerciseService.class);

    private final ApplicationProperties applicationProperties;

    private final FileService fileService;

    private final GitlabService gitlabService;

    public ExerciseService(ApplicationProperties applicationProperties, FileService fileService, GitlabService gitlabService) {
        this.applicationProperties = applicationProperties;
        this.fileService = fileService;
        this.gitlabService = gitlabService;
    }

    /**
     * Retrieve a zipped Artemis Exercise from a given URL
     * and store it in 'tmp-import' directory
     * @param exerciseUrl the url of the exposed exercise
     * @throws ArtemisImportError
     */
    public void importExerciseFromArtemis(String exerciseUrl) throws ArtemisImportError{
        String exerciseToken = exerciseUrl.split("/")[exerciseUrl.split("/").length - 1];
        Client client = ClientBuilder.newClient();

        WebTarget target = client.target(exerciseUrl);
        try {
            InputStream inputStream = target.request().accept(javax.ws.rs.core.MediaType.APPLICATION_OCTET_STREAM).get(InputStream.class);

            ZipInputStream zipInput = new ZipInputStream(inputStream);
            storeZipInTmp(zipInput, exerciseToken);

        } catch (Exception e) {
            throw new ArtemisImportError("ZipFile couldn't be retrieved from [" + URLDecoder.decode(exerciseUrl, StandardCharsets.UTF_8) + "]" + e.getMessage());
        }
    }

    /**
     * Import exercise to Gitlab.
     * Sequentially executes following processes:
     * - Apply user defined exercise metadata
     * - Create repository in Gitlab's group defined by the user
     * - Locally checking out the new created repository
     * - Copy, commit and push files to the new repository
     *
     * @param exerciseInfo of the exercise to import
     * @param exerciseToken of the stored exercise (in '/tmp-import' directory) to import
     */
    public void importExercise(ArtemisExerciseInfo exerciseInfo, String exerciseToken, Integer gitlabGroupId) throws GitLabApiException, GitAPIException, IOException, ArtemisImportError {
        applyExerciseInfoChanges(exerciseInfo, exerciseToken);
        renameFiles(exerciseToken);

        gitlabService.setJGitDefaultAuth();
        Group group = gitlabService.getGroupById(gitlabGroupId);
        String repoUrl = gitlabService.createRepository(group, exerciseInfo.getTitle());
        Repository repo = gitlabService.checkoutRepo(repoUrl);
        gitlabService.commitAndPushToRepo(repo, new File("tmp-import", exerciseToken));

        repo.close();
        try {
            FileUtils.deleteDirectory(new File("tmp-import", exerciseToken));
            FileUtils.deleteDirectory(repo.getDirectory().getParentFile());
        } catch (IOException e) {
            log.error("Could not remove directory: {} \n", e.getMessage());
        }
    }

    /**
     * Retrieves metadata from imported Artemis exercise
     * @param exerciseToken of the imported exercise
     * @return Content of exercise's 'metadata.yaml' and 'artemis.yaml'
     */
    public ArtemisExerciseInfo getArtemisExerciseInfo(String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new JsonFactory());
        mapper.enable(DeserializationFeature.READ_ENUMS_USING_TO_STRING)
            .enable(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS)
            .enable(DeserializationFeature.ACCEPT_SINGLE_VALUE_AS_ARRAY)
            .disable(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES);;
        File exerciseDetailsFile = Objects.requireNonNull(new File("tmp-import", exerciseToken).listFiles((file, name) -> name.startsWith("Exercise-Details-") && name.endsWith(".json")))[0];

        mapArtemisExerciseInfoValues(exerciseDetailsFile);

        ArtemisExerciseInfo artemisExerciseInfo = mapper.readValue(exerciseDetailsFile, ArtemisExerciseInfo.class);
        writeArtemisMetadataFile(exerciseDetailsFile, artemisExerciseInfo);

        return artemisExerciseInfo;
    }

    /**
     * Utility method used to map Artemis Exercise metadata values to the ones used in
     * Sharing Metadata
     *
     * @param jsonFile to parse the values from
     * @throws IOException if an error occurs
     */
    private void mapArtemisExerciseInfoValues(File jsonFile) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        JSONObject jsonObject = mapper.readValue(jsonFile, JSONObject.class);
        if (jsonObject.containsKey("difficulty")) {
            switch (jsonObject.get("difficulty").toString()) {
                case "EASY":
                    jsonObject.put("difficulty", "SIMPLE");
                case "MEDIUM":
                    jsonObject.put("difficulty", "MEDIUM");
                case "HARD":
                    jsonObject.put("difficulty", "ADVANCED");
            }
        }

        if (!jsonObject.containsKey("metadataVersion")) {
            jsonObject.put("metadataVersion", "0.4");
        }

        if (!jsonObject.containsKey("license")) {
            jsonObject.put("license", "CC-SA-BY 4.0");
        }

        if (!jsonObject.containsKey("format")) {
            jsonObject.put("format", "artemis");
        }

        if (!jsonObject.containsKey("structure")) {
            jsonObject.put("structure", "atomic");
        }

        if (jsonObject.containsKey("instructors")) {
            jsonObject.put("creator", jsonObject.get("instructors"));
            jsonObject.remove("instructors");
        }

        if ("programming".equals(jsonObject.get("type").toString())) {
            jsonObject.put("type", "programming exercise");
        }

        FileWriter fileWriter = new FileWriter(jsonFile);
        fileWriter.write(jsonObject.toString(JSONStyle.NO_COMPRESS));
        fileWriter.close();
    }

    /**
     * Utility method used to extract Artemis metadata into a 'artemis.yaml' file
     * @param jsonMetadataFile where the exercise metadata is stored
     * @param artemisExerciseInfo the entity of the stored metadata
     * @throws IOException if an error occurs
     */
    private void writeArtemisMetadataFile(File jsonMetadataFile, ArtemisExerciseInfo artemisExerciseInfo) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory()
            .enable(YAMLGenerator.Feature.INDENT_ARRAYS)
            .disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER));

        File artemisMetadataFile = new File(jsonMetadataFile.getParent(), "artemis.yaml");
        if (!artemisMetadataFile.exists()) {
            artemisMetadataFile.createNewFile();
        }
        mapper.writeValue(artemisMetadataFile, artemisExerciseInfo.getArtemisInfoOnly());
    }

    /**
     * Used to temporary store content of Zip-File into a tmp-subdirectory
     *
     * Recursion is used in case of nested Zip-Files
     *
     * @param zipInputStream containing the Zip-File to unzip
     * @param subDir name of the tmp-subdirectory
     */
    private void storeZipInTmp(ZipInputStream zipInputStream, String subDir) throws IOException {
        if (!new File("tmp-import").exists() && !new File("tmp-import").mkdir()) {
            throw new IOException("Could not create missing '/tmp-import' directory");
        }
        File tmpDir = new File("tmp-import", subDir);
        if (!tmpDir.exists() && !tmpDir.mkdir()) {
            throw new IOException("Temporary exercise directory could not be created");
        }

        ZipEntry zipEntry;

        // https://www.baeldung.com/java-compress-and-uncompress
        while ((zipEntry = zipInputStream.getNextEntry()) != null) {
            if (zipEntry.getName().contains("/.git/")) {    // important as Artemis exercises come in submodules
                continue;
            }
            File entryFile = new File(tmpDir, zipEntry.getName());
            if (zipEntry.getName().endsWith(".zip")) {
                writeFile(entryFile, zipInputStream);
                String newSubDir;
                if (StringUtils.endsWithAny(zipEntry.getName(), new String[]{"-exercise.zip", "-solution.zip", "-tests.zip"})) {
                    // extract repo directory name only
                    newSubDir = Paths.get(entryFile.getParentFile().getParent(), zipEntry.getName().substring(zipEntry.getName().lastIndexOf("-") + 1, zipEntry.getName().lastIndexOf("."))).toString();
                } else {
                    newSubDir = Paths.get(subDir, entryFile.getName()).toString();
                    newSubDir = newSubDir.substring(0, newSubDir.lastIndexOf("."));     // remove '.zip' from path
                }
                if (newSubDir.startsWith("tmp-import")) {
                    newSubDir = newSubDir.replaceFirst("tmp-import/", "");
                }
                storeZipInTmp(new ZipInputStream(new FileInputStream(entryFile)), newSubDir);
                if (Files.deleteIfExists(entryFile.toPath())) {
                    log.warn("Unable to delete temporary Zip file [{}] after extraction. Skipping..", entryFile.getPath());
                }
            }
            else if (zipEntry.isDirectory()) {
                if (!entryFile.mkdir()) {
                    throw new IOException("Failed to create directory " + entryFile);
                }
            } else if (!zipEntry.getName().endsWith(".zip")){
                // fix for Windows-created archives
                File parent = entryFile.getParentFile();
                if (!parent.isDirectory() && !parent.mkdirs()) {
                    throw new IOException("Failed to create directory " + parent);
                }

                writeFile(entryFile, zipInputStream);
            }
        }
        if (tmpDir.isDirectory() && fileService.isEmpty(tmpDir.toPath())){
            Files.delete(tmpDir.toPath());
        }
        zipInputStream.close();
        fileService.scheduleForDeletion(tmpDir.toPath(), 60);
    }

    /**
     * Method used to write files from zip inputStreams
     * @param entryFile File to write
     * @param zipInputStream Byte stream to write
     * @throws IOException
     */
    private void writeFile(File entryFile, ZipInputStream zipInputStream) throws IOException {
        byte[] buffer = new byte[1024];

        FileOutputStream fos = new FileOutputStream(entryFile);
        int len;
        while ((len = zipInputStream.read(buffer)) > 0) {
            fos.write(buffer, 0, len);
        }
        fos.close();
    }

    /**
     * Utility method used to apply metadata changes to the exercise's 'metadata.yaml' file
     * @param exerciseInfo containing the metadata values
     * @param exerciseToken of the exercise to update
     * @throws IOException
     */
    public void applyExerciseInfoChanges(ArtemisExerciseInfo exerciseInfo, String exerciseToken) throws IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory()
            .enable(YAMLGenerator.Feature.INDENT_ARRAYS)
            .disable(YAMLGenerator.Feature.WRITE_DOC_START_MARKER));
        mapper.writeValue(new File("tmp-import/" + exerciseToken, "metadata.yaml"), exerciseInfo.getExerciseInfoOnly());
    }

    /**
     * Used to validate apiKeys by checking if they match with
     * one of the registeredConnectors' keys stored in application-(dev|prod).yml
     * @param apiKey the apikey to validate
     * @return true if equal to the local one, false otherwise
     */
    public boolean validate(String apiKey) {
        for (ApplicationProperties.RegisteredConnector registeredConnector : applicationProperties.getRegisteredConnectors()) {
            if (registeredConnector.getAccessToken().equals(apiKey)) {
                log.debug("ApiKey validated correctly against [" + registeredConnector.getUrl() + "]");
                return true;
            }
        }
        return false;
    }

    /**
     * Renames specific files of an imported Artemis exercise,
     * to meet naming conventions
     *
     * Renamed files:
     * - Problem-Statement-xxx.md -> exercise.md
     *
     * @param exerciseToken to retrieve the location of the files to rename
     */
    private void renameFiles(String exerciseToken) {
        try {
            File problemStatementFile = Objects.requireNonNull(new File("tmp-import", exerciseToken).listFiles((file, name) -> name.startsWith("Problem-Statement-") && name.endsWith(".md")))[0];
            if (!problemStatementFile.renameTo(new File("tmp-import" + File.separator + exerciseToken, "exercise.md"))) {
                log.error("Unable to rename file [{}]. Skipping..", problemStatementFile.getName());
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            log.error("File to rename could not be found in [{}].", "tmp-import" + File.separator + exerciseToken);
        }

    }
}
