import { Component, OnInit } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { ExerciseService } from 'app/exercise/service/exercise.service';
import { ArtemisExerciseInfo } from 'app/shared/model/artemis-exercise-info.model';
import { JhiAlertService } from 'ng-jhipster';
import { Person } from 'app/shared/model/person.model';
import { GitlabService } from 'app/shared/gitlab/gitlab.service';
import { NgbModal } from '@ng-bootstrap/ng-bootstrap';
import { GitlabPathSelectorComponent } from 'app/shared/gitlab/gitlab-path-selector/gitlab-path-selector.component';
import { GitlabGroup } from 'app/shared/model/gitlab-group.model';

@Component({
  selector: 'jhi-import',
  templateUrl: './exercise-import.component.html',
  styleUrls: ['./exercise-import.component.scss'],
})
export class ExerciseImportComponent implements OnInit {
  exerciseToken: string | undefined;
  exerciseInfo: ArtemisExerciseInfo | undefined;

  languages = ['english', 'german'];
  difficulties = ['simple', 'medium', 'advanced'];

  newCreator = { name: '', email: '', affiliation: '' } as Person;
  newPublisher = { name: '', email: '', affiliation: '' } as Person;
  isPublisherSameAsCreator = true;

  selectedGitlabGroup: GitlabGroup | undefined;
  isLoading = false;
  isImporting = false;

  constructor(
    private route: ActivatedRoute,
    private exerciseService: ExerciseService,
    private jhiAlertService: JhiAlertService,
    private gitlabService: GitlabService,
    private modalService: NgbModal
  ) {}

  ngOnInit() {
    this.isLoading = true;
    this.exerciseToken = this.route.snapshot.params['token'];
    if (this.exerciseToken) {
      this.exerciseService.importExerciseInfoFromArtemis(this.exerciseToken).subscribe(
        data => {
          this.exerciseInfo = data;
          this.exerciseInfo.programmingLanguage = data['programmingLanguage'];
          this.exerciseInfo.programmingLanguage = this.exerciseInfo.programmingLanguage.map(programmingLanguage =>
            this.toTitleCase(programmingLanguage)
          );
          this.exerciseInfo.creator = data['creator'];
          if (this.exerciseInfo.language === undefined || this.exerciseInfo.language === null || this.exerciseInfo.language.length === 0) {
            this.exerciseInfo.language = ['english']; // default value if none is set
          }
          if (this.exerciseInfo.identifier === undefined || this.exerciseInfo.identifier.localeCompare('')) {
            this.route.queryParams.subscribe(params => {
              this.exerciseInfo!.identifier = params['callback'];
            });
          }
          if (this.exerciseInfo.learningResourceType === undefined || this.exerciseInfo.learningResourceType.localeCompare('')) {
            this.exerciseInfo.learningResourceType = 'Artemis';
          }
          if (this.exerciseInfo.keyword === undefined || this.exerciseInfo.keyword.length < 1) {
            this.exerciseInfo.keyword = ['artemis'];
          }
        },
        error => {
          this.jhiAlertService.error('', { error: error.error });
        }
      );
    }
    this.isLoading = false;
  }

  /**
   * Actual import function submitting import form to server
   */
  import() {
    if (this.exerciseInfo!.difficulty !== null) {
      this.exerciseInfo!.difficulty = this.exerciseInfo!.difficulty?.toLowerCase();
    }
    if (this.isPublisherSameAsCreator) {
      this.exerciseInfo!.publisher = this.exerciseInfo?.creator!;
    }
    this.openGitlabPathSelectorModal().then(gitlabGroup => {
      if (gitlabGroup !== undefined) {
        this.isLoading = true;
        this.isImporting = true;
        this.exerciseService.submitExerciseInfoForImport(this.exerciseInfo!, this.exerciseToken!, gitlabGroup.id).subscribe(
          () => {
            window.location.href = this.route.snapshot.queryParams['callback'];
          },
          error => {
            this.jhiAlertService.error(error.error.detail);
            window.scroll(0, 0);
            this.isLoading = false;
            this.isImporting = false;
          }
        );
      }
    });
    this.isLoading = false;
    this.isImporting = false;
  }

  /**
   * Utility function used to open the Gitlab selector modal
   */
  openGitlabPathSelectorModal() {
    const modalRef = this.modalService.open(GitlabPathSelectorComponent, { size: 'xl', backdrop: 'static' });
    return modalRef.result.then(group => {
      this.selectedGitlabGroup = group;
      return group;
    });
  }

  /**
   * Validation function for button-groups.
   * At least one value must be set.
   * @return true if invalid and false if valid
   */
  hasValidationErrors() {
    if (
      this.exerciseInfo?.keyword === undefined ||
      this.exerciseInfo?.language === undefined ||
      !(this.exerciseInfo.keyword.length > 0) ||
      !(this.exerciseInfo.language.length > 0)
    ) {
      return true;
    }
    return false;
  }

  /**
   * Used to set the exercise's difficulty on
   * user's click
   * @param difficulty to set
   */
  setDifficulty(difficulty: string) {
    this.exerciseInfo!.difficulty = difficulty;
  }

  /**
   * Used to add a language to the exercise
   * @param language to add
   */
  setLanguage(language: string) {
    if (this.exerciseInfo!.language.includes(language)) {
      this.exerciseInfo!.language = this.exerciseInfo!.language.filter(lang => lang !== language);
    } else {
      this.exerciseInfo!.language.push(language);
    }
  }

  /**
   * Used to add a new person to a given list of persons
   * @param person to add
   * @param personList to add the given person to
   */
  addPerson(person: Person, personList: Person[]) {
    if (this.validatePerson(person)) {
      if (this.exerciseInfo && personList === undefined) {
        personList = [];
      }
      personList.push(person);

      // needed in order to modify person object by reference
      person.name = '';
      person.email = '';
      person.affiliation = '';
    }
  }

  /**
   * Used to remove a person from a given list of persons
   * @param personToRemove to remove
   * @param personList to remove the given person from
   */
  removePerson(personToRemove: Person, personList: Person[]) {
    personList.filter(person => person !== personToRemove);
  }

  /**
   * Validates person by checking if all attributes are set
   *
   * @param person to check
   * @private
   */
  private validatePerson(person: Person) {
    if (person.name === undefined || person.affiliation === undefined || person.email === undefined) {
      return false;
    }
    if (person.name === '' || person.affiliation === '' || person.email === '') {
      return false;
    }
    return true;
  }

  /**
   * Utility function used to turn words into title case
   * @param word the word to turn into title case
   */
  private toTitleCase(word: string) {
    return word[0].toUpperCase() + word.substr(1).toLowerCase();
  }

  /**
   * Utility function used to check if the given difficulty is
   * the same as the imported exercise
   * @param difficulty to check
   */
  isDifficulty(difficulty: string) {
    if (this.exerciseInfo!.difficulty !== undefined) {
      return this.exerciseInfo!.difficulty.toLowerCase() === difficulty.toLowerCase();
    } else {
      return false;
    }
  }

  /**
   * Utility function used to check if a given language is set
   * in the imported exercsie
   * @param language
   */
  isLanguage(language: string) {
    return this.exerciseInfo!.language !== undefined && this.exerciseInfo!.language.includes(language);
  }

  /**
   * Used to validate language input
   */
  isLanguageSet() {
    return this.exerciseInfo!.language !== undefined && this.exerciseInfo!.language.length > 0;
  }

  /**
   * Checks if at least one person is set for the given list of persons
   */
  isPersonSet(personList: Person[]) {
    return personList !== undefined && personList.length > 0;
  }
}
